#include <common.h>
#include <ppc_asm.tmpl>
#include <ppc4xx.h>
#include <asm/processor.h>
#include <exports.h>
#include "config.h"
#include "flash_local.h"

DECLARE_GLOBAL_DATA_PTR;

extern flash_info_t  flash_info[]; /* info for FLASH chips */
extern unsigned long __dummy;

void do_reset (void);
int do_updater(void);
int flash_print_info(flash_info_t * info);
static ulong flash_get_size(vu_long * addr, flash_info_t * info);

pci_dev_t pci_find_devices(struct pci_device_id *ids, int index)
{
	struct pci_controller * hose;
	u16 vendor, device;
	u8 header_type;
	pci_dev_t bdf;
	int i, bus, found_multi = 0;

	for (hose = gd->ppc440_hose; hose; hose = hose->next)
	{
		for (bus = hose->first_busno; bus <= hose->last_busno; bus++)
		{
			for (bdf = PCI_BDF(bus,0,0); bdf < PCI_BDF(bus+1,0,0); bdf += PCI_BDF(0,0,1))
			{
				if (!PCI_FUNC(bdf)) {
					hose->read_byte(hose, bdf, PCI_HEADER_TYPE, &header_type);

					found_multi = header_type & 0x80;
				} else {
					if (!found_multi)
						continue;
				}

				hose->read_word(hose, bdf, PCI_VENDOR_ID,&vendor);
				hose->read_word(hose, bdf,PCI_DEVICE_ID,&device);

				for (i=0; ids[i].vendor != 0; i++)
					if (vendor == ids[i].vendor &&
					    device == ids[i].device)
					{
						if (index <= 0)
							return bdf;

						index--;
					}
			}
		}
	}

	return (-1);
}

pci_dev_t pci_find_device(unsigned int vendor, unsigned int device, int index)
{
	static struct pci_device_id ids[2] = {{}, {0, 0}};

	ids[0].vendor = vendor;
	ids[0].device = device;

	return pci_find_devices(ids, index);
}

void _main(void)
{
	pci_dev_t dev;

	printf("\nU-Boot Firmware Updater\n\n");

	if ( ! gd->ppc440_hose) return;

    // check if we're running on a Sam-flex, Sam-flex-se or Sam-mini-itx

#ifdef CONFIG_SAM440EP
	if ((dev = pci_find_device(0x1002,0x4c66,0)) == -1) // ATI Radeon M9 not present
	{
	    printf("****************************************************\n");
		printf("* ATTENTION!! THIS UPDATER IS FOR SAM-MINI ONLY !! *\n");
		printf("****************************************************\n\n");

		printf("Please reset the board !!!\n\n");

		while (1);
	}
#endif
#ifdef CONFIG_SAM440EP_FLEX
	if (((dev = pci_find_device(0x1002,0x4c66,0)) >= 0) || // ATI Radeon M9
		((dev = pci_find_device(0x1131,0x1561,0)) >= 0)) // ISP1561 present
	{
	    printf("****************************************************\n");
		printf("* ATTENTION!! THIS UPDATER IS FOR SAM-FLEX ONLY !! *\n");
		printf("****************************************************\n\n");

		printf("Please reset the board !!!\n\n");

		while (1);
	}
#endif
#ifdef CONFIG_SAM440EP_FLEX_SE
	if (((dev = pci_find_device(0x1002,0x4c66,0)) >= 0) || // ATI Radeon M9
		((dev = pci_find_device(0x1033,0x0035,0)) >= 0)) // NEC present
	{
	    printf("*******************************************************\n");
		printf("* ATTENTION!! THIS UPDATER IS FOR SAM-FLEX-SE ONLY !! *\n");
		printf("*******************************************************\n\n");

		printf("Please reset the board !!!\n\n");

		while (1);
	}
#endif

	printf("****************************************************\n");
	printf("*  ATTENTION!! PLEASE READ THIS NOTICE CAREFULLY!  *\n");
	printf("****************************************************\n\n");
	printf("This program  will update your computer's  firmware.\n");
	printf("Do NOT  remove the disk,  reset the  machine,  or do\n");
	printf("anything that  might disrupt functionality.  If this\n");
	printf("Program fails, your computer  might be unusable, and\n");
	printf("you will  need to return your  board for reflashing.\n");
	printf("If you find this  too  risky,  remove the cdrom  and\n");
	printf("switch off your  machine now.  Otherwise  press the \n");
	printf("SPACE key now to start the process\n\n");

	do
	{
		char x;
		while (!tstc());
		x = getc();
		if (x == ' ') break;
		if (x == 'q') return;
	} while (1);

	int rc = do_updater();

	if (rc == 0)
	{
		printf("\nUpdate done. Please remove the cdrom.\n");
		printf("You can switch off/reset now when the cdrom is removed\n\n");
	}
	printf("Press the R key to reset\n\n");

	do
	{
		char x;
		while (!tstc());
		x = getc();
		if (x == 'r') do_reset();
	} while (1);
}

int do_updater(void)
{
	unsigned long *addr = &__dummy + 65;
	int rc = 0;

	flash_get_size(0xfff80000,&flash_info[0]);
	rc = flash_print_info(&flash_info[0]);
	if (rc != 0) printf("Error: you need to use another updater with this FLASH chip\n");
	else
	{
		flash_sect_protect(0, 0xFFF80000, 0xFFFFFFFF);

		printf("\nErasing ");
		flash_sect_erase(0xFFF80000, 0xFFFFFFFF);
		printf("Writing ");
		rc = flash_write((uchar *)addr, 0xFFF80000, 0x80000);
		if (rc != 0) printf(" Flashing failed due to error %d\n", rc);
		else printf(" done\n");

		flash_sect_protect(1, 0xFFF80000, 0xFFFFFFFF);
	}

	return rc;
}
